/*
 * lhuff.h - simple Huffman coding and other routines for storing data
 *
 * Copyright (c) 2008, Petteri Kaski, Patric Ostergard, Olli Pottonen
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include "lhuff.h"
#include "compression.h"

/******************************************************* Read and write bits */

static unsigned int  readbit(bb_in *b);
static void          writebit(bb_out *b, unsigned char bit);

static unsigned int readbit(bb_in *b)
{
    if(b->bitcount == 0) {
        int c;
        b->bitcount = CHAR_BIT;
        b->bits = c = fgetc(b->f);
        if(c == EOF)
            stsc_error("readbit(): read failed\n");
    }

    unsigned int res = b->bits & 1U;
    b->bits >>= 1;
    --b->bitcount;
    return res;
}

static void writebit(bb_out *b, unsigned char bit)
{
    b->bits |= (bit << b->bitcount);

    if(++b->bitcount == CHAR_BIT) {
        if(fputc(b->bits, b->f) == EOF) 
            stsc_error("writebit(): write failed\n");
        b->bits     = 0;
        b->bitcount = 0;
    }
}

/************************************************************* Initalization */

void init_bb_in(bb_in *bb, FILE *f)
{
    bb->f        = f;
    bb->bits     = 0;
    bb->bitcount = 0;
}

void init_bb_out(bb_out *bb, FILE *f)
{
    bb->f        = f;
    bb->bits     = 0;
    bb->bitcount = 0;
}

/********************************************************** Flush the buffer */

void flush_bb_in(bb_in *b)
{
    b->bitcount = 0;
}

void flush_bb_out(bb_out *b)
{
    if(b->bitcount == 0)
        return; /* Nothing to flush */

    b->bitcount = 0;
    if(fputc(b->bits, b->f) == EOF)
        stsc_error("flushbb(): write failed\n");
    b->bits = 0;
}

/************************************************* Seeking positions of file */

void bb_in_seek(bb_in *r, long pos)
{
    r->bits = r->bitcount = 0;
    if(fseek(r->f, pos, SEEK_SET) == -1)
        stsc_error("fseek() failed\n");
}

void bb_out_seek(bb_out *w, long pos)
{
    w->bits = w->bitcount = 0;
    if(fseek(w->f, pos, SEEK_SET) == -1)
        stsc_error("fseek() failed\n");
}

/************************************************** Current position in file */

long bb_in_pos(const bb_in *r)
{
    long p = ftell(r->f);
    if(p == -1)
        stsc_error("ftell() fails\n");
    return p;
}

long bb_out_pos(const bb_out *w)
{
    long p = ftell(w->f);
    if(p == -1)
        stsc_error("ftell() fails\n");
    return p;
}

/******************************************************* Read and write ints */

uint32_t readint(bb_in *r)
{
    int b1, b2, b3, b4;
    b1 = fgetc(r->f);
    b2 = fgetc(r->f);
    b3 = fgetc(r->f);
    b4 = fgetc(r->f);
    if(b1 == EOF || b2 == EOF || b3 == EOF || b4 == EOF)
        stsc_error("readint(): read failed\n");
    
    uint32_t a = (((uint32_t)b1) << 24) |
        (((uint32_t)b2) << 16) |
        (((uint32_t)b3) << 8)  |
        ((uint32_t)b4);
    return a;
}

void writeint(bb_out *w, uint32_t a)
{
    int error = (fputc((a >> 24) & 0xff, w->f) == EOF ||
                 fputc((a >> 16) & 0xff, w->f) == EOF ||
                 fputc( (a >> 8) & 0xff, w->f) == EOF ||
                 fputc(        a & 0xff, w->f) == EOF);
    if(error)
        stsc_error("writeint(): write failed\n");
}

/************************************************************ Huffman coding */

/* Maximum number of symbols, length of codewords and number of nodes in
 * the Huffman tree. */
#define MAXSYMBOLS     (STS_B + 1)
#define MAXCWLEN       (MAXSYMBOLS-1)
#define MAX_NODE_COUNT (2*MAXSYMBOLS - 1)

typedef struct h_node huffman_node;
struct h_node {
    unsigned char isLeaf;
    union {
        huffman_node * children[2];
        unsigned char  symbol;
    } chldr;
};

typedef struct {
    unsigned long len;
    unsigned char bits[MAXCWLEN];
} huffman_codeword;

/* Struct for a huffman code. It might be slightly more efficient to have two
 * different structs, one for encoding and another for decoding. */
struct h_code {
    /* nodes[0] is the root node */
    huffman_node     nodes[MAX_NODE_COUNT];
    huffman_codeword codewords[MAXSYMBOLS];
    int symbolcount;
};

/*********************************************** Constructing a Huffman code */

/* Transform a string to a codeword, return the length. */
static int init_codeword(huffman_codeword *cw, const char *cw_str)
{
    for(int i= 0; i < MAXCWLEN; i++)
        switch(*cw_str++) {
            case '0':
                cw->bits[i] = 0; break;
            case '1':
                cw->bits[i] = 1; break;
            case '\0':
                return i;
            default:
                stsc_error("init_code(): invalid codeword\n");
        }

    stsc_error("init_code(): too long codeword\n");
    /* Unreachable */
    return -1;
}

static void init_path(huffman_code *c, const huffman_codeword *cw, int *nodecount,
                     int symbol)
{
    huffman_node *p = &c->nodes[0]; /* root node */
    int cwlen = cw->len;
    /* Check if root node is still uninitialized. Note that the root node may
     * be a leaf, if the code consists of only one string,
     * namely the empty string. */
    if(p->isLeaf == UCHAR_MAX)
        p->isLeaf = (cwlen == 0);

    for(int i = 0; i < cwlen; i++) {
        int bit              = cw->bits[i];
        huffman_node **child = &p->chldr.children[bit];

        if(p->isLeaf)
            stsc_error("init_code(): codeword is prefix of another codeword\n");
        if(*child == NULL) {
            /* Allocate & initialize a new node */
            if(*nodecount == MAX_NODE_COUNT)
                stsc_error("init_code(): redundant code\n");
            *child                      = &c->nodes[(*nodecount)++];
            (*child)->isLeaf            = (i+1 == cwlen);
            (*child)->chldr.children[0] = NULL;
            (*child)->chldr.children[1] = NULL;
        }
        p = *child;
    }
    if(!p->isLeaf)
        stsc_error("init_code(): codeword is prefix of another codeword\n");
    p->chldr.symbol = symbol;
}

/* The argument shall be an array of at most MAXSYMBOL codewords
 * followed by NULL, represented as e.g. "010".  */
huffman_code *init_code(const char * const *codewords)
{
    huffman_code *c = malloc(sizeof(huffman_code));
    if(c == NULL)
        stsc_error("init_code(): malloc() failed\n");
    memset(c, 0, sizeof(huffman_code));

    /* Count the symbols */
    int symbolcount = 0;
    while(codewords[symbolcount] != NULL)
        symbolcount++;
    if(symbolcount > MAXSYMBOLS)
        stsc_error("init_code(): too many codewords\n");

    /* Initialize the codewords */
    for(int i = 0; i < symbolcount; ++i)
        c->codewords[i].len = init_codeword(&c->codewords[i], codewords[i]);

    /* Partially initialize root node */
    c->nodes[0].isLeaf = UCHAR_MAX; /* special value for unitiliazide root */
    c->nodes[0].chldr.children[0] = NULL;
    c->nodes[0].chldr.children[1] = NULL;
    int nodecount = 1;

    /* Initialize the Huffman tree by initializing path to each leaf */
    for(int i = 0; i < symbolcount; i++)
        init_path(c, &c->codewords[i], &nodecount, i);

    /* Check that each node has two children or is a leaf */
    for(int i = 0; i < nodecount; i++) {
        huffman_node *n = &c->nodes[i];
        if(n->isLeaf == 0 &&
           (n->chldr.children[0] == NULL || n->chldr.children[1] == NULL))
            stsc_error("init_code(): invalid code\n");
    }

    return c;
}

/********************************************************** Release function */

void free_code(huffman_code *c)
{
    free(c);
}

/********************************************************* Decoding a symbol */

int decode_symbol(const huffman_code *c, bb_in *in)
{
    const huffman_node *p = &c->nodes[0];
    while(!p->isLeaf)
        p = p->chldr.children[readbit(in)];

    return p->chldr.symbol;
}

/********************************************************* Encoding a symbol */

void encode_symbol(const huffman_code *c, bb_out *out, int symbol)
{
    const huffman_codeword *codew = &c->codewords[symbol];
    for(unsigned int i = 0; i < codew->len; i++)
        writebit(out, codew->bits[i]);
}

/********************************************* Print error message and abort */

void stsc_error(const char *msg)
{
    fputs(msg, stderr);
    abort();
}
